/*
 * Paging object coordination.
 *
 * Copyright (C) 2018, 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/c/util/cap_alloc.h>
#include <l4/re/env.h>
#include <l4/sys/factory.h>

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "paging.h"



Paging::Paging(Pages *pages) : _pages(pages)
{
  _map_semaphore = l4re_util_cap_alloc();
  
  if (l4_is_valid_cap(_map_semaphore))
  { 
    l4_factory_create(l4re_env()->factory, L4_PROTO_SEMAPHORE, _map_semaphore);
    l4_semaphore_up(_map_semaphore);
  }
}

/* Look up any existing object for the file. */

Accessor *Paging::find(unsigned long fileid)
{
  Accessor_map::iterator accessor_it = _accessor_map.find(fileid);

  if (accessor_it == _accessor_map.end())
    return 0;

  printf("Obtain existing accessor for %ld: %p\n", fileid, accessor_it->second);

  return accessor_it->second;
}

/* Remove any existing object for the file. */

void Paging::remove(unsigned long fileid)
{
  l4_semaphore_down(_map_semaphore, L4_IPC_NEVER);

  Accessor_map::iterator accessor_it = _accessor_map.find(fileid);

  if (accessor_it != _accessor_map.end())
  {
    _accessor_map.erase(accessor_it);

    printf("Removed accessor for %ld!\n", fileid);
  }
  else
    printf("Could not remove accessor for %ld!\n", fileid);

  l4_semaphore_up(_map_semaphore);
}

/* Set an object for access to the file. */

Accessor *Paging::set_conditional(unsigned long fileid, Accessor *accessor)
{
  Accessor *found_accessor;

  l4_semaphore_down(_map_semaphore, L4_IPC_NEVER);

  found_accessor = find(fileid);

  if (!found_accessor)
  {
    _accessor_map[fileid] = accessor;
    found_accessor = accessor;

    printf("Set accessor for %ld: %p\n", fileid, accessor);
  }
  else
    printf("Retrieved accessor for %ld: %p\n", fileid, found_accessor);

  found_accessor->attach();

  l4_semaphore_up(_map_semaphore);

  return found_accessor;
}
